"""Various constants"""
AUTH0_URL = "https://login.sharkninja.com"
AUTH0_HOST = "login.sharkninja.com"
AUTH0_CLIENT_ID = "wsguxrqm77mq4LtrTrwg8ZJUxmSrexGi"
AUTH0_SCOPES = "openid profile email offline_access"
AUTH0_REDIRECT_URI = "com.sharkninja.shark://login.sharkninja.com/ios/com.sharkninja.shark/callback"
DEVICE_URL = "https://ads-field-39a9391a.aylanetworks.com"
LOGIN_URL = "https://user-field-39a9391a.aylanetworks.com"
SHARK_APP_ID = "ios_shark_prod-3A-id"
SHARK_APP_SECRET = "ios_shark_prod-74tFWGNg34LQCmR0m45SsThqrqs"
SHARK_APP_USERAGENT = "SharkClean/29562 Darwin/24.3.0"
BROWSER_USERAGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/135.0.0.0 Safari/537.36"
EU_AUTH0_URL = "https://logineu.sharkninja.com"
EU_AUTH0_HOST = "logineu.sharkninja.com"
EU_AUTH0_CLIENT_ID = "rKDx9O18dBrY3eoJMTkRiBZHDvd9Mx1I"
EU_DEVICE_URL = "https://ads-eu.aylanetworks.com"
EU_LOGIN_URL = "https://user-field-eu.aylanetworks.com"
EU_SHARK_APP_ID = "android_shark_prod-lg-id"
EU_SHARK_APP_SECRET = "android_shark_prod-xuf9mlHOo0p3Ty5bboFROSyRBlE"
